/* global DOMParser */
var matchesSelector = require('./matchesSelector');

function getJSDOM() {
  try {
    return require('' + 'jsdom');
  } catch (err) {
    throw new Error(
      'unexpected-dom: Running outside a browser (or in a browser without DOMParser), but could not find the `jsdom` module. Please npm install jsdom to make this work.'
    );
  }
}

function getHtmlDocument(str) {
  if (typeof DOMParser !== 'undefined') {
    return new DOMParser().parseFromString(str, 'text/html');
  } else if (
    typeof document !== 'undefined' &&
    document.implementation &&
    document.implementation.createHTMLDocument
  ) {
    var htmlDocument = document.implementation.createHTMLDocument('');
    htmlDocument.open();
    htmlDocument.write(str);
    htmlDocument.close();
    return htmlDocument;
  } else {
    var jsdom = getJSDOM();

    return jsdom.JSDOM
      ? new jsdom.JSDOM(str).window.document
      : jsdom.jsdom(str);
  }
}

function parseHtml(str, isFragment) {
  if (isFragment) {
    str = "<html><head></head><body>" + str + "</body></html>";
  }
  var htmlDocument = getHtmlDocument(str);

  if (isFragment) {
    var body = htmlDocument.body;
    var documentFragment = htmlDocument.createDocumentFragment();
    if (body) {
      for (var i = 0; i < body.childNodes.length; i += 1) {
        documentFragment.appendChild(body.childNodes[i].cloneNode(true));
      }
    }
    return documentFragment;
  } else {
    return htmlDocument;
  }
}

function parseXml(str) {
  if (typeof DOMParser !== 'undefined') {
    return new DOMParser().parseFromString(str, 'text/xml');
  } else {
    var jsdom = getJSDOM();

    if (jsdom.JSDOM) {
      return new jsdom.JSDOM(str, { contentType: 'text/xml' }).window.document;
    } else {
      return jsdom.jsdom(str, { parsingMode: 'xml' });
    }
  }
}

// From html-minifier
var enumeratedAttributeValues = {
  draggable: ['true', 'false'] // defaults to 'auto'
};

var matchSimpleAttribute = /^(?:allowfullscreen|async|autofocus|autoplay|checked|compact|controls|declare|default|defaultchecked|defaultmuted|defaultselected|defer|disabled|enabled|formnovalidate|hidden|indeterminate|inert|ismap|itemscope|loop|multiple|muted|nohref|noresize|noshade|novalidate|nowrap|open|pauseonexit|readonly|required|reversed|scoped|seamless|selected|sortable|spellcheck|truespeed|typemustmatch|visible)$/i;

function isBooleanAttribute(attrName) {
  return matchSimpleAttribute.test(attrName);
}

function isEnumeratedAttribute(attrName) {
  return attrName in enumeratedAttributeValues;
}

function validateStyles(expect, str) {
  var invalidStyles = str
    .split(';')
    .filter(
      function (part) { return !/^\s*(\w|-)+\s*:\s*(#(?:[0-9a-fA-F]{3}){1,2}|[^#]+)\s*$|^$/.test(part); }
    );

  if (invalidStyles.length > 0) {
    expect.errorMode = 'nested';
    expect.fail(
      'Expectation contains invalid styles: {0}',
      invalidStyles.join(';')
    );
  }
}

function styleStringToObject(str) {
  var styles = {};

  str.split(';').forEach(function (rule) {
    var colonIndex = rule.indexOf(':');

    // Guard against empty touples
    if (colonIndex !== -1) {
      var key = rule.slice(0, colonIndex).trim();
      var value = rule.slice(colonIndex + 1).trim();

      styles[key] = value;
    }
  });

  return styles;
}

function getClassNamesFromAttributeValue(attributeValue) {
  if (attributeValue === null) {
    return [];
  }

  if (attributeValue === '') {
    return [];
  }

  var classNames = attributeValue.split(/\s+/);
  if (classNames.length === 1 && classNames[0] === '') {
    classNames.pop();
  }
  return classNames;
}

function isInsideHtmlDocument(node) {
  var ownerDocument =
    node.nodeType === 9 && node.documentElement && node.implementation
      ? node
      : node.ownerDocument;

  if (ownerDocument.contentType) {
    return ownerDocument.contentType === 'text/html';
  } else {
    return ownerDocument.toString() === '[object HTMLDocument]';
  }
}

function getAttributes(element) {
  var isHtml = isInsideHtmlDocument(element);
  var attrs = element.attributes;
  var result = {};

  for (var i = 0; i < attrs.length; i += 1) {
    if (attrs[i].name === 'class') {
      result[attrs[i].name] =
        (attrs[i].value && attrs[i].value.split(' ')) || [];
    } else if (attrs[i].name === 'style') {
      result[attrs[i].name] = styleStringToObject(attrs[i].value);
    } else {
      result[attrs[i].name] =
        isHtml && isBooleanAttribute(attrs[i].name)
          ? true
          : attrs[i].value || '';
    }
  }

  return result;
}

function entitify(value) {
  return String(value)
    .replace(/&/g, '&amp;')
    .replace(/"/g, '&quot;')
    .replace(/</g, '&lt;');
}

function isVoidElement(elementName) {
  return /(?:area|base|br|col|embed|hr|img|input|keygen|link|menuitem|meta|param|source|track|wbr)/i.test(
    elementName
  );
}

function writeAttributeToMagicPen(output, attributeName, value, isHtml) {
  output.prismAttrName(attributeName);
  if (!isHtml || !isBooleanAttribute(attributeName)) {
    if (attributeName === 'class') {
      value = value.join(' ');
    } else if (attributeName === 'style') {
      value = Object.keys(value)
        .map(function (cssProp) { return (cssProp + ": " + (value[cssProp])); })
        .join('; ');
    }
    output
      .prismPunctuation('="')
      .prismAttrValue(entitify(value))
      .prismPunctuation('"');
  }
}

function stringifyAttribute(attributeName, value) {
  if (
    isBooleanAttribute(attributeName) ||
    isEnumeratedAttribute(attributeName)
  ) {
    return attributeName;
  } else if (attributeName === 'class') {
    return ("class=\"" + (value.join(' ')) + "\""); // FIXME: entitify
  } else if (attributeName === 'style') {
    return ("style=\"" + (Object.keys(value)
      // FIXME: entitify
      .map(function (cssProp) { return [cssProp, value[cssProp]].join(': '); })
      .join('; ')) + "\"");
  } else {
    return (attributeName + "=\"" + (entitify(value)) + "\"");
  }
}

function stringifyStartTag(element) {
  var elementName = isInsideHtmlDocument(element)
    ? element.nodeName.toLowerCase()
    : element.nodeName;
  var str = "<" + elementName;
  var attrs = getAttributes(element);

  Object.keys(attrs).forEach(function (key) {
    str += " " + (stringifyAttribute(key, attrs[key]));
  });

  str += '>';
  return str;
}

function stringifyEndTag(element) {
  var isHtml = isInsideHtmlDocument(element);
  var elementName = isHtml
    ? element.nodeName.toLowerCase()
    : element.nodeName;
  if (isHtml && isVoidElement(elementName) && element.childNodes.length === 0) {
    return '';
  } else {
    return ("</" + elementName + ">");
  }
}

function ensureSupportedSpecOptions(options) {
  var unsupportedOptions = Object.keys(options).filter(
    function (key) { return key !== 'attributes' &&
      key !== 'name' &&
      key !== 'children' &&
      key !== 'onlyAttributes' &&
      key !== 'textContent'; }
  );

  if (unsupportedOptions.length > 0) {
    throw new Error(
      ("Unsupported option" + (unsupportedOptions.length === 1 ? '' : 's') + ": " + (unsupportedOptions.join(', ')))
    );
  }
}

module.exports = {
  name: 'unexpected-dom',
  installInto: function installInto(expect) {
    expect = expect.child();
    expect.use(require('magicpen-prism'));

    function bubbleError(body) {
      return expect.withError(body, function (err) {
        err.errorMode = 'bubble';
        throw err;
      });
    }

    expect.exportType({
      name: 'DOMNode',
      base: 'object',
      identify: function identify(obj) {
        return (
          obj &&
          obj.nodeName &&
          [2, 3, 4, 5, 6, 7, 10, 11, 12].indexOf(obj.nodeType) > -1
        );
      },
      equal: function equal(a, b) {
        return a.nodeType === b.nodeType && a.nodeValue === b.nodeValue;
      },
      diff: function () {},
      inspect: function inspect(element, depth, output) {
        return output.code(
          ((element.nodeName) + " \"" + (element.nodeValue) + "\""),
          'prism-string'
        );
      }
    });

    expect.exportType({
      name: 'DOMComment',
      base: 'DOMNode',
      identify: function identify(obj) {
        return obj && typeof obj.nodeType === 'number' && obj.nodeType === 8;
      },
      equal: function equal(a, b) {
        return a.nodeValue === b.nodeValue;
      },
      inspect: function inspect(element, depth, output) {
        return output.code(("<!--" + (element.nodeValue) + "-->"), 'html');
      },
      diff: function diff(actual, expected, output, diff$1, inspect, equal) {
        var d = diff$1(
          ("<!--" + (actual.nodeValue) + "-->"),
          ("<!--" + (expected.nodeValue) + "-->")
        );
        d.inline = true;
        return d;
      }
    });

    // Recognize <!-- ignore --> as a special subtype of DOMComment so it can be targeted by assertions:
    expect.exportType({
      name: 'DOMIgnoreComment',
      base: 'DOMComment',
      identify: function identify(obj) {
        return (
          this.baseType.identify(obj) && /^\s*ignore\s*$/.test(obj.nodeValue)
        );
      }
    });

    expect.exportType({
      name: 'DOMTextNode',
      base: 'DOMNode',
      identify: function identify(obj) {
        return obj && typeof obj.nodeType === 'number' && obj.nodeType === 3;
      },
      equal: function equal(a, b) {
        return a.nodeValue === b.nodeValue;
      },
      inspect: function inspect(element, depth, output) {
        return output.code(entitify(element.nodeValue.trim()), 'html');
      },
      diff: function diff(actual, expected, output, diff$1, inspect, equal) {
        var d = diff$1(actual.nodeValue, expected.nodeValue);
        d.inline = true;
        return d;
      }
    });

    expect.exportType({
      name: 'DOMNodeList',
      base: 'array-like',
      prefix: function prefix(output) {
        return output.text('NodeList[');
      },
      suffix: function suffix(output) {
        return output.text(']');
      },
      similar: function similar(a, b) {
        // Figure out whether a and b are "struturally similar" so they can be diffed inline.
        return (
          a.nodeType === 1 && b.nodeType === 1 && a.nodeName === b.nodeName
        );
      },
      identify: function identify(obj) {
        return (
          obj &&
          typeof obj.length === 'number' &&
          typeof obj.toString === 'function' &&
          typeof obj.item === 'function' &&
          // With jsdom 6+, nodeList.toString() comes out as '[object Object]', so fall back to the constructor name:
          (obj.toString().indexOf('NodeList') !== -1 ||
            (obj.constructor && obj.constructor.name === 'NodeList'))
        );
      }
    });

    // Fake type to make it possible to build 'to satisfy' diffs to be rendered inline:
    expect.exportType({
      name: 'attachedDOMNodeList',
      base: 'DOMNodeList',
      indent: false,
      prefix: function prefix(output) {
        return output;
      },
      suffix: function suffix(output) {
        return output;
      },
      delimiter: function delimiter(output) {
        return output;
      },
      identify: function identify(obj) {
        return obj && obj._isAttachedDOMNodeList;
      }
    });

    function makeAttachedDOMNodeList(domNodeList, contentType) {
      var attachedDOMNodeList = [];
      for (var i = 0; i < domNodeList.length; i += 1) {
        attachedDOMNodeList.push(domNodeList[i]);
      }
      attachedDOMNodeList._isAttachedDOMNodeList = true;
      attachedDOMNodeList.ownerDocument = { contentType: contentType };
      return attachedDOMNodeList;
    }

    expect.exportType({
      name: 'HTMLDocType',
      base: 'DOMNode',
      identify: function identify(obj) {
        return (
          obj &&
          typeof obj.nodeType === 'number' &&
          obj.nodeType === 10 &&
          'publicId' in obj
        );
      },
      inspect: function inspect(doctype, depth, output, inspect$1) {
        return output.code(("<!DOCTYPE " + (doctype.name) + ">"), 'html');
      },
      equal: function equal(a, b) {
        return a.toString() === b.toString();
      },
      diff: function diff(actual, expected, output, diff$1) {
        var d = diff$1(
          ("<!DOCTYPE " + (actual.name) + ">"),
          ("<!DOCTYPE " + (expected.name) + ">")
        );
        d.inline = true;
        return d;
      }
    });

    expect.exportType({
      name: 'DOMDocument',
      base: 'DOMNode',
      identify: function identify(obj) {
        return (
          obj &&
          typeof obj.nodeType === 'number' &&
          obj.nodeType === 9 &&
          obj.documentElement &&
          obj.implementation
        );
      },
      inspect: function inspect(document, depth, output, inspect$1) {
        for (var i = 0; i < document.childNodes.length; i += 1) {
          output.append(inspect$1(document.childNodes[i]));
        }
        return output;
      },
      equal: function equal(a, b, equal$1) {
        return equal$1(a.childNodes, b.childNodes);
      },
      diff: function diff(actual, expected, output, diff$1, inspect, equal) {
        output.inline = true;
        output.append(
          diff$1(
            makeAttachedDOMNodeList(actual.childNodes),
            makeAttachedDOMNodeList(expected.childNodes)
          )
        );
        return output;
      }
    });

    expect.exportType({
      name: 'HTMLDocument',
      base: 'DOMDocument',
      identify: function identify(obj) {
        return this.baseType.identify(obj) && isInsideHtmlDocument(obj);
      }
    });

    expect.exportType({
      name: 'XMLDocument',
      base: 'DOMDocument',
      identify: function identify(obj) {
        return this.baseType.identify(obj) && !isInsideHtmlDocument(obj);
      },
      inspect: function inspect(document, depth, output, inspect$1) {
        output.code('<?xml version="1.0"?>', 'xml');
        for (var i = 0; i < document.childNodes.length; i += 1) {
          output.append(inspect$1(document.childNodes[i], depth - 1));
        }
        return output;
      }
    });

    expect.exportType({
      name: 'DOMDocumentFragment',
      base: 'DOMNode',
      identify: function identify(obj) {
        return obj && obj.nodeType === 11; // In jsdom, documentFragment.toString() does not return [object DocumentFragment]
      },
      inspect: function inspect(documentFragment, depth, output, inspect$1) {
        return output
          .text('DocumentFragment[')
          .append(inspect$1(documentFragment.childNodes, depth))
          .text(']');
      },
      equal: function equal(a, b, equal$1) {
        return equal$1(a.childNodes, b.childNodes);
      },
      diff: function diff(actual, expected, output, diff$1, inspect, equal) {
        output.inline = true;
        output.block(
          diff$1(
            makeAttachedDOMNodeList(actual.childNodes),
            makeAttachedDOMNodeList(expected.childNodes)
          )
        );
        return output;
      }
    });

    expect.exportType({
      name: 'DOMElement',
      base: 'DOMNode',
      identify: function identify(obj) {
        return (
          obj &&
          typeof obj.nodeType === 'number' &&
          obj.nodeType === 1 &&
          obj.nodeName &&
          obj.attributes
        );
      },
      equal: function equal(a, b, equal$1) {
        var aIsHtml = isInsideHtmlDocument(a);
        var bIsHtml = isInsideHtmlDocument(b);
        return (
          aIsHtml === bIsHtml &&
          (aIsHtml
            ? a.nodeName.toLowerCase() === b.nodeName.toLowerCase()
            : a.nodeName === b.nodeName) &&
          equal$1(getAttributes(a), getAttributes(b)) &&
          equal$1(a.childNodes, b.childNodes)
        );
      },
      inspect: function inspect(element, depth, output, inspect$1) {
        var elementName = element.nodeName.toLowerCase();
        var startTag = stringifyStartTag(element);

        output.code(startTag, 'html');
        if (element.childNodes.length > 0) {
          if (depth === 1) {
            output.text('...');
          } else {
            var inspectedChildren = [];
            if (elementName === 'script') {
              var type = element.getAttribute('type');
              if (!type || /javascript/.test(type)) {
                type = 'javascript';
              }
              inspectedChildren.push(
                output.clone().code(element.textContent, type)
              );
            } else if (elementName === 'style') {
              inspectedChildren.push(
                output
                  .clone()
                  .code(
                    element.textContent,
                    element.getAttribute('type') || 'text/css'
                  )
              );
            } else {
              for (var i = 0; i < element.childNodes.length; i += 1) {
                inspectedChildren.push(inspect$1(element.childNodes[i]));
              }
            }

            var width = startTag.length;
            var sizes = inspectedChildren.map(function (inspectedChild) { return inspectedChild.size(); }
            );

            var multipleLines = sizes.some(function (size) {
              width += size.width;
              return width > 60 || size.height > 1;
            });

            if (multipleLines) {
              output.nl().indentLines();

              inspectedChildren.forEach(function (inspectedChild, index) {
                var childSize = sizes[index];
                if (childSize.width > 0 && childSize.height > 0) {
                  output.i().block(inspectedChild);
                }

                output.nl();
              });

              output.outdentLines();
            } else {
              inspectedChildren.forEach(function (inspectedChild, index) { return output.append(inspectedChild); }
              );
            }
          }
        }
        output.code(stringifyEndTag(element), 'html');
        return output;
      },
      diffLimit: 512,
      diff: function diff(actual, expected, output, diff$1, inspect, equal) {
        var isHtml = isInsideHtmlDocument(actual);
        output.inline = true;

        if (Math.max(actual.length, expected.length) > this.diffLimit) {
          output.jsComment(("Diff suppressed due to size > " + (this.diffLimit)));
          return output;
        }

        var emptyElements =
          actual.childNodes.length === 0 && expected.childNodes.length === 0;
        var conflictingElement =
          actual.nodeName.toLowerCase() !== expected.nodeName.toLowerCase() ||
          !equal(getAttributes(actual), getAttributes(expected));

        if (conflictingElement) {
          var canContinueLine = true;
          output.prismPunctuation('<').prismTag(actual.nodeName.toLowerCase());
          if (
            actual.nodeName.toLowerCase() !== expected.nodeName.toLowerCase()
          ) {
            output
              .sp()
              .annotationBlock(function (output) { return output
                  .error('should be')
                  .sp()
                  .prismTag(expected.nodeName.toLowerCase()); }
              )
              .nl();
            canContinueLine = false;
          }
          var actualAttributes = getAttributes(actual);
          var expectedAttributes = getAttributes(expected);
          Object.keys(actualAttributes).forEach(function (attributeName) {
            output.sp(canContinueLine ? 1 : 2 + actual.nodeName.length);
            writeAttributeToMagicPen(
              output,
              attributeName,
              actualAttributes[attributeName],
              isHtml
            );
            if (attributeName in expectedAttributes) {
              if (
                actualAttributes[attributeName] ===
                expectedAttributes[attributeName]
              ) {
                canContinueLine = true;
              } else {
                output
                  .sp()
                  .annotationBlock(function (output) { return output
                      .error('should equal')
                      .sp()
                      .append(
                        inspect(entitify(expectedAttributes[attributeName]))
                      ); }
                  )
                  .nl();
                canContinueLine = false;
              }
              delete expectedAttributes[attributeName];
            } else {
              output
                .sp()
                .annotationBlock(function (output) { return output.error('should be removed'); })
                .nl();
              canContinueLine = false;
            }
          });
          Object.keys(expectedAttributes).forEach(function (attributeName) {
            output.sp(canContinueLine ? 1 : 2 + actual.nodeName.length);
            output
              .annotationBlock(function (output) {
                output.error('missing').sp();
                writeAttributeToMagicPen(
                  output,
                  attributeName,
                  expectedAttributes[attributeName],
                  isHtml
                );
              })
              .nl();
            canContinueLine = false;
          });
          output.prismPunctuation('>');
        } else {
          output.code(stringifyStartTag(actual), 'html');
        }

        if (!emptyElements) {
          output
            .nl()
            .indentLines()
            .i()
            .block(
              diff$1(
                makeAttachedDOMNodeList(actual.childNodes),
                makeAttachedDOMNodeList(expected.childNodes)
              )
            )
            .nl()
            .outdentLines();
        }

        output.code(stringifyEndTag(actual), 'html');
        return output;
      }
    });

    expect.exportAssertion(
      '<DOMElement> to have (class|classes) <array|string>',
      function (expect, subject, value) { return expect(subject, 'to have attributes', { class: value }); }
    );

    expect.exportAssertion(
      '<DOMElement> to only have (class|classes) <array|string>',
      function (expect, subject, value) { return expect(subject, 'to have attributes', {
          class: expect.it(function (className) {
            var actualClasses = getClassNamesFromAttributeValue(className);
            if (typeof value === 'string') {
              value = getClassNamesFromAttributeValue(value);
            }
            return bubbleError(function () { return expect(actualClasses.sort(), 'to equal', value.sort()); }
            );
          })
        }); }
    );

    expect.exportAssertion(
      '<DOMElement> not to have (class|classes) <array|string>',
      function (expect, subject, value) {
        return expect(subject, 'to have attributes', {
          class: expect.it(function (className) {
            var actualClasses = getClassNamesFromAttributeValue(className);
            var expectedClasses;
            if (typeof value === 'string') {
              expectedClasses = getClassNamesFromAttributeValue(value);
            } else {
              expectedClasses = value;
            }
            return bubbleError(function () { return expect.apply(void 0, [ actualClasses, 'not to contain' ].concat( expectedClasses )); }
            );
          })
        });
      }
    );

    expect.exportAssertion(
      '<DOMTextNode> to [exhaustively] satisfy <DOMTextNode>',
      function (expect, subject, value) { return expect(subject.nodeValue, 'to equal', value.nodeValue); }
    );

    expect.exportAssertion(
      '<DOMComment> to [exhaustively] satisfy <DOMComment>',
      function (expect, subject, value) { return expect(subject.nodeValue, 'to equal', value.nodeValue); }
    );

    // Avoid rendering a huge object diff when a text node is matched against a different node type:
    expect.exportAssertion(
      '<DOMTextNode> to [exhaustively] satisfy <object>',
      function (expect, subject, value) { return expect.fail(); }
    );

    // Always passes:
    expect.exportAssertion(
      // Name each subject type to increase the specificity of the assertion
      '<DOMComment|DOMElement|DOMTextNode|DOMDocument|HTMLDocType> to [exhaustively] satisfy <DOMIgnoreComment>',
      function (expect, subject, value) {}
    );

    // Necessary because this case would otherwise be handled by the above catch-all for <object>:
    expect.exportAssertion(
      '<DOMTextNode> to [exhaustively] satisfy <regexp>',
      function (expect, ref, value) {
        var nodeValue = ref.nodeValue;

        return expect(nodeValue, 'to satisfy', value);
    }
    );

    expect.exportAssertion(
      '<DOMTextNode> to [exhaustively] satisfy <any>',
      function (expect, ref, value) {
        var nodeValue = ref.nodeValue;

        return expect(nodeValue, 'to satisfy', value);
    }
    );

    function convertDOMNodeToSatisfySpec(node, isHtml) {
      if (node.nodeType === 10) {
        // HTMLDocType
        return { name: node.nodeName };
      } else if (node.nodeType === 1) {
        // DOMElement
        var name = isHtml ? node.nodeName.toLowerCase() : node.nodeName;

        var result = { name: name };

        if (node.attributes) {
          result.attributes = {};
          for (var i = 0; i < node.attributes.length; i += 1) {
            result.attributes[node.attributes[i].name] =
              isHtml && isBooleanAttribute(node.attributes[i].name)
                ? true
                : node.attributes[i].value || '';
          }
        }
        result.children = Array.prototype.slice.call(node.childNodes);
        return result;
      } else if (node.nodeType === 3) {
        // DOMTextNode
        return node.nodeValue;
      } else if (node.nodeType === 8) {
        // DOMComment
        return node;
      } else {
        throw new Error(
          ("to satisfy: Node type " + (node.nodeType) + " is not yet supported in the value")
        );
      }
    }

    expect.exportAssertion(
      '<DOMNodeList> to [exhaustively] satisfy <string>',
      function (expect, subject, value) {
        var isHtml = isInsideHtmlDocument(subject);

        expect.argsOutput = function (output) { return output.code(value, isHtml ? 'html' : 'xml'); };

        return expect(
          subject,
          'to [exhaustively] satisfy',
          (isHtml ? parseHtml(value, true) : parseXml(value)).childNodes
        );
      }
    );

    expect.exportAssertion(
      '<DOMNodeList> to [exhaustively] satisfy <DOMNodeList>',
      function (expect, subject, value) {
        var isHtml = isInsideHtmlDocument(subject);
        var satisfySpecs = [];
        for (var i = 0; i < value.length; i += 1) {
          satisfySpecs.push(convertDOMNodeToSatisfySpec(value[i], isHtml));
        }
        return expect(subject, 'to [exhaustively] satisfy', satisfySpecs);
      }
    );

    expect.exportAssertion(
      '<DOMDocumentFragment> to [exhaustively] satisfy <string>',
      function (expect, subject, value) {
        var isHtml = isInsideHtmlDocument(subject);

        expect.argsOutput = function (output) { return output.code(value, isHtml ? 'html' : 'xml'); };

        return expect(
          subject,
          'to [exhaustively] satisfy',
          isHtml ? parseHtml(value, true) : parseXml(value)
        );
      }
    );

    expect.exportAssertion(
      '<DOMDocumentFragment> to [exhaustively] satisfy <DOMDocumentFragment>',
      function (expect, subject, ref) {
        var childNodes = ref.childNodes;

        var isHtml = isInsideHtmlDocument(subject);
        return expect(
          subject,
          'to [exhaustively] satisfy',
          Array.prototype.map.call(childNodes, function (childNode) { return convertDOMNodeToSatisfySpec(childNode, isHtml); }
          )
        );
      }
    );

    expect.exportAssertion(
      '<DOMDocumentFragment> to [exhaustively] satisfy <object|array>',
      function (expect, ref, value) {
          var childNodes = ref.childNodes;

          return expect(childNodes, 'to [exhaustively] satisfy', value);
    }
    );

    expect.exportAssertion(
      '<DOMElement> to [exhaustively] satisfy <string>',
      function (expect, subject, value) {
        var isHtml = isInsideHtmlDocument(subject);
        var documentFragment = isHtml
          ? parseHtml(value, true)
          : parseXml(value);
        if (documentFragment.childNodes.length !== 1) {
          throw new Error(
            'HTMLElement to satisfy string: Only a single node is supported'
          );
        }

        expect.argsOutput = function (output) { return output.code(value, isHtml ? 'html' : 'xml'); };

        return expect(
          subject,
          'to [exhaustively] satisfy',
          documentFragment.childNodes[0]
        );
      }
    );

    expect.exportAssertion(
      '<DOMDocument> to [exhaustively] satisfy <string>',
      function (expect, subject, value) {
        var isHtml = isInsideHtmlDocument(subject);
        var valueDocument = isHtml
          ? parseHtml(value, false)
          : parseXml(value);
        return expect(
          makeAttachedDOMNodeList(subject.childNodes),
          'to [exhaustively] satisfy',
          Array.prototype.map.call(valueDocument.childNodes, function (childNode) { return convertDOMNodeToSatisfySpec(childNode, isHtml); }
          )
        );
      }
    );

    expect.exportAssertion(
      '<DOMDocument> to [exhaustively] satisfy <DOMDocument>',
      function (expect, subject, ref) {
        var childNodes = ref.childNodes;

        var isHtml = isInsideHtmlDocument(subject);
        return expect(
          makeAttachedDOMNodeList(subject.childNodes),
          'to [exhaustively] satisfy',
          Array.prototype.map.call(childNodes, function (childNode) { return convertDOMNodeToSatisfySpec(childNode, isHtml); }
          )
        );
      }
    );

    expect.exportAssertion(
      '<DOMElement> to [exhaustively] satisfy <DOMElement>',
      function (expect, subject, value) { return expect(
          subject,
          'to [exhaustively] satisfy',
          convertDOMNodeToSatisfySpec(value, isInsideHtmlDocument(subject))
        ); }
    );

    expect.exportAssertion(
      [
        '<DOMElement> to [exhaustively] satisfy <DOMDocumentFragment>',
        '<DOMElement> to [exhaustively] satisfy <DOMTextNode>',
        '<DOMTextNode> to [exhaustively] satisfy <DOMElement>',
        '<DOMElement|DOMDocumentFragment|DOMDocument> to [exhaustively] satisfy <regexp>',
        '<DOMDocumentFragment|DOMDocument> to [exhaustively] satisfy <DOMElement>'
      ],
      function (expect, subject, value) { return expect.fail(); }
    );

    expect.exportAssertion(
      '<DOMElement> to [exhaustively] satisfy <object>',
      function (expect, subject, value) {
        var isHtml = isInsideHtmlDocument(subject);

        if (expect.argTypes[0].is('expect.it')) {
          expect.context.thisObject = subject;
          return value(subject, expect.context);
        }

        ensureSupportedSpecOptions(value);
        var promiseByKey = {
          name: expect.promise(function () {
            if (value && typeof value.name !== 'undefined') {
              return bubbleError(function () { return expect(
                  isHtml ? subject.nodeName.toLowerCase() : subject.nodeName,
                  'to satisfy',
                  value.name
                ); }
              );
            }
          }),
          children: expect.promise(function () {
            if (typeof value.children !== 'undefined') {
              if (typeof value.textContent !== 'undefined') {
                throw new Error(
                  'The children and textContent properties are not supported together'
                );
              }

              var contentType = subject.ownerDocument.contentType;
              if (!contentType) {
                // provide a value in the absence of a contentType (IE11)
                contentType = isInsideHtmlDocument(subject)
                  ? 'text/html'
                  : 'application/xml';
              }

              return bubbleError(function () { return expect(
                  makeAttachedDOMNodeList(subject.childNodes, contentType),
                  'to satisfy',
                  value.children
                ); }
              );
            } else if (typeof value.textContent !== 'undefined') {
              return bubbleError(function () { return expect(subject.textContent, 'to satisfy', value.textContent); }
              );
            }
          }),
          attributes: {}
        };

        var onlyAttributes =
          (value && value.onlyAttributes) || expect.flags.exhaustively;
        var attrs = getAttributes(subject);
        var expectedAttributes = value && value.attributes;
        var expectedAttributeNames = [];
        var expectedValueByAttributeName = {};

        if (typeof expectedAttributes !== 'undefined') {
          if (typeof expectedAttributes === 'string') {
            expectedAttributes = [expectedAttributes];
          }
          if (Array.isArray(expectedAttributes)) {
            expectedAttributes.forEach(function (attributeName) {
              expectedValueByAttributeName[attributeName] = true;
            });
          } else if (
            expectedAttributes &&
            typeof expectedAttributes === 'object'
          ) {
            expectedValueByAttributeName = expectedAttributes;
          }
          Object.keys(expectedValueByAttributeName).forEach(function (attributeName) {
            expectedAttributeNames.push(attributeName);
          });

          expectedAttributeNames.forEach(function (attributeName) {
            var attributeValue = subject.getAttribute(attributeName);
            var expectedAttributeValue =
              expectedValueByAttributeName[attributeName];
            promiseByKey.attributes[attributeName] = expect.promise(function () {
              if (typeof expectedAttributeValue === 'undefined') {
                return bubbleError(function () { return expect(subject.hasAttribute(attributeName), 'to be false'); }
                );
              } else if (isEnumeratedAttribute(attributeName)) {
                var indexOfEnumeratedAttributeValue = enumeratedAttributeValues[
                  attributeName
                ].indexOf(expectedAttributeValue);

                return bubbleError(function () {
                  if (indexOfEnumeratedAttributeValue === -1) {
                    expect.fail(function (output) { return output
                        .text('Invalid expected value ')
                        .appendInspected(expectedAttributeValue)
                        .text('. Supported values include: ')
                        .appendItems(
                          enumeratedAttributeValues[attributeName],
                          ', '
                        ); }
                    );
                  }

                  expect(attributeValue, 'to satisfy', expectedAttributeValue);
                });
              } else if (expectedAttributeValue === true) {
                return bubbleError(function () { return expect(subject.hasAttribute(attributeName), 'to be true'); }
                );
              } else if (
                attributeName === 'class' &&
                (typeof expectedAttributeValue === 'string' ||
                  Array.isArray(expectedAttributeValue))
              ) {
                var actualClasses = getClassNamesFromAttributeValue(
                  attributeValue
                );
                var expectedClasses = expectedAttributeValue;
                if (typeof expectedClasses === 'string') {
                  expectedClasses = getClassNamesFromAttributeValue(
                    expectedAttributeValue
                  );
                }
                if (onlyAttributes) {
                  return bubbleError(function () { return expect(
                      actualClasses.sort(),
                      'to equal',
                      expectedClasses.sort()
                    ); }
                  );
                } else {
                  if (expectedClasses.length === 0) {
                    return bubbleError(function () { return expect(expectedClasses, 'to be empty'); }
                    );
                  }
                  return bubbleError(function () { return expect.apply(void 0, [ actualClasses, 'to contain' ].concat( expectedClasses )); }
                  );
                }
              } else if (attributeName === 'style') {
                var expectedStyleObj;
                if (typeof expectedValueByAttributeName.style === 'string') {
                  validateStyles(expect, expectedValueByAttributeName.style);
                  expectedStyleObj = styleStringToObject(
                    expectedValueByAttributeName.style
                  );
                } else {
                  expectedStyleObj = expectedValueByAttributeName.style;
                }

                if (onlyAttributes) {
                  return bubbleError(function () { return expect(
                      attrs.style,
                      'to exhaustively satisfy',
                      expectedStyleObj
                    ); }
                  );
                } else {
                  return bubbleError(function () { return expect(attrs.style, 'to satisfy', expectedStyleObj); }
                  );
                }
              } else if (
                expect.findTypeOf(expectedAttributeValue).is('expect.it')
              ) {
                expect.context.thisObject = subject;
                return bubbleError(function () { return expectedAttributeValue(attributeValue, expect.context); }
                );
              } else {
                return bubbleError(function () { return expect(attributeValue, 'to satisfy', expectedAttributeValue); }
                );
              }
            });
          });

          promiseByKey.attributePresence = expect.promise(function () {
            var attributeNamesExpectedToBeDefined = [];
            expectedAttributeNames.forEach(function (attributeName) {
              if (
                typeof expectedValueByAttributeName[attributeName] ===
                'undefined'
              ) {
                expect(attrs, 'not to have key', attributeName);
              } else {
                attributeNamesExpectedToBeDefined.push(attributeName);
                expect(attrs, 'to have key', attributeName);
              }
            });
            if (onlyAttributes) {
              expect(
                Object.keys(attrs).sort(),
                'to equal',
                attributeNamesExpectedToBeDefined.sort()
              );
            }
          });
        }

        return expect.promise.all(promiseByKey).caught(function () { return expect.promise.settle(promiseByKey).then(function () {
            expect.fail({
              diff: function diff(output, diff$1, inspect, equal) {
                output.block(function (output) {
                  var seenError = false;
                  output
                    .prismPunctuation('<')
                    .prismTag(
                      isHtml ? subject.nodeName.toLowerCase() : subject.nodeName
                    );
                  if (promiseByKey.name.isRejected()) {
                    seenError = true;
                    var nameError = promiseByKey.name.reason();
                    output.sp().annotationBlock(function (output) { return output
                        .error(
                          (nameError && nameError.getLabel()) ||
                            'should satisfy'
                        )
                        .sp()
                        .append(inspect(value.name)); }
                    );
                  }
                  var inspectedAttributes = [];
                  Object.keys(attrs).forEach(function (attributeName) {
                    var attributeOutput = output.clone();
                    var promise = promiseByKey.attributes[attributeName];
                    writeAttributeToMagicPen(
                      attributeOutput,
                      attributeName,
                      attrs[attributeName],
                      isHtml
                    );
                    if (
                      (promise && promise.isFulfilled()) ||
                      (!promise &&
                        (!onlyAttributes ||
                          expectedAttributeNames.indexOf(attributeName) !== -1))
                    ) {
                    } else {
                      seenError = true;
                      attributeOutput.sp().annotationBlock(function (output) {
                        if (
                          promise &&
                          typeof expectedValueByAttributeName[attributeName] !==
                            'undefined'
                        ) {
                          output.appendErrorMessage(promise.reason());
                        } else {
                          // onlyAttributes === true
                          output.error('should be removed');
                        }
                      });
                    }
                    inspectedAttributes.push(attributeOutput);
                  });
                  expectedAttributeNames.forEach(function (attributeName) {
                    if (!subject.hasAttribute(attributeName)) {
                      var promise = promiseByKey.attributes[attributeName];
                      if (!promise || promise.isRejected()) {
                        seenError = true;
                        var err = promise && promise.reason();
                        var attributeOutput = output
                          .clone()
                          .annotationBlock(function (output) {
                            output
                              .error('missing')
                              .sp()
                              .prismAttrName(attributeName, 'html');
                            if (
                              expectedValueByAttributeName[attributeName] !==
                              true
                            ) {
                              output
                                .sp()
                                .error(
                                  (err && err.getLabel()) || 'should satisfy'
                                )
                                .sp()
                                .append(
                                  inspect(
                                    expectedValueByAttributeName[attributeName]
                                  )
                                );
                            }
                          });
                        inspectedAttributes.push(attributeOutput);
                      }
                    }
                  });
                  if (inspectedAttributes.length > 0) {
                    if (seenError) {
                      output
                        .nl()
                        .indentLines()
                        .indent()
                        .block(function (output) {
                          inspectedAttributes.forEach(function (item, i) {
                            if (i > 0) {
                              output.nl();
                            }
                            output.append(item);
                          });
                        })
                        .outdentLines()
                        .nl();
                    } else {
                      output.sp();
                      inspectedAttributes.forEach(function (item, i) {
                        if (i > 0) {
                          output.sp();
                        }
                        output.append(item);
                      });
                    }
                  } else if (seenError) {
                    // The tag name mismatched
                    output.nl();
                  }

                  output.prismPunctuation('>');
                  var childrenError =
                    promiseByKey.children.isRejected() &&
                    promiseByKey.children.reason();
                  if (childrenError) {
                    var childrenDiff = childrenError.getDiff(output);
                    if (childrenDiff && childrenDiff.inline) {
                      output
                        .nl()
                        .indentLines()
                        .i()
                        .block(childrenDiff)
                        .nl()
                        .outdentLines();
                    } else {
                      output
                        .nl()
                        .indentLines()
                        .i()
                        .block(function (output) {
                          for (
                            var i = 0;
                            i < subject.childNodes.length;
                            i += 1
                          ) {
                            output.append(inspect(subject.childNodes[i])).nl();
                          }
                        });
                      output
                        .sp()
                        .annotationBlock(function (output) { return output.appendErrorMessage(childrenError); }
                        );
                      output.nl();
                    }
                  } else {
                    for (var i = 0; i < subject.childNodes.length; i += 1) {
                      output.append(inspect(subject.childNodes[i]));
                    }
                  }
                  output.code(stringifyEndTag(subject), 'html');
                });
                output.inline = true;
                return output;
              }
            });
          }); }
        );
      }
    );

    expect.exportAssertion(
      '<DOMElement> to [only] have (attribute|attributes) <string+>',
      function (expect, subject) {
          var args = [], len = arguments.length - 2;
          while ( len-- > 0 ) args[ len ] = arguments[ len + 2 ];

          return expect(subject, 'to [only] have attributes', args);
    }
    );

    expect.exportAssertion(
      '<DOMElement> not to have (attribute|attributes) <array>',
      function (expect, subject, value) {
        var attributes = getAttributes(subject);

        value.forEach(function (name) {
          delete attributes[name];
        });

        return expect(subject, 'to only have attributes', attributes);
      }
    );

    expect.exportAssertion(
      '<DOMElement> not to have (attribute|attributes) <string+>',
      function (expect, subject) {
          var args = [], len = arguments.length - 2;
          while ( len-- > 0 ) args[ len ] = arguments[ len + 2 ];

          return expect(subject, 'not to have attributes', args);
    }
    );

    expect.exportAssertion(
      '<DOMElement> to [only] have (attribute|attributes) <array|object>',
      function (expect, subject, value) { return expect(subject, 'to satisfy', {
          attributes: value,
          onlyAttributes: expect.flags.only
        }); }
    );

    expect.exportAssertion(
      '<DOMElement> to have [no] (child|children)',
      function (expect, ref) {
          var childNodes = ref.childNodes;

          return expect.flags.no
          ? expect(childNodes, 'to be empty')
          : expect(childNodes, 'not to be empty');
    }
    );

    expect.exportAssertion(
      '<DOMElement> to have text <any>',
      function (expect, ref, value) {
          var textContent = ref.textContent;

          return expect(textContent, 'to satisfy', value);
    }
    );

    expect.exportAssertion(
      '<DOMDocument|DOMElement|DOMDocumentFragment> [when] queried for [first] <string> <assertion?>',
      function (expect, subject, query) {
        var queryResult;

        expect.argsOutput[0] = function (output) { return output.green(query); };
        expect.errorMode = 'nested';

        if (expect.flags.first) {
          queryResult = subject.querySelector(query);
          if (!queryResult) {
            expect.subjectOutput = function (output) { return expect.inspect(subject, Infinity, output); };

            expect.fail(function (output) { return output
                .error('The selector')
                .sp()
                .jsString(query)
                .sp()
                .error('yielded no results'); }
            );
          }
        } else {
          queryResult = subject.querySelectorAll(query);
          if (queryResult.length === 0) {
            expect.subjectOutput = function (output) { return expect.inspect(subject, Infinity, output); };

            expect.fail(function (output) { return output
                .error('The selector')
                .sp()
                .jsString(query)
                .sp()
                .error('yielded no results'); }
            );
          }
        }
        return expect.shift(queryResult);
      }
    );

    expect.exportAssertion(
      '<DOMDocument|DOMElement|DOMDocumentFragment> [when] queried for test id <string> <assertion?>',
      function (expect, subject, testId) {
        expect.errorMode = 'nested';

        var escapedTestId = JSON.stringify(testId);

        return expect(
          subject,
          'queried for first',
          ("[data-test-id=" + escapedTestId + "]")
        ).then(function (queryResult) { return expect.shift(queryResult); });
      }
    );

    expect.exportAssertion(
      [
        '<DOMDocument|DOMElement|DOMDocumentFragment> to contain [no] elements matching <string>',
        '<DOMDocument|DOMElement|DOMDocumentFragment> [not] to contain elements matching <string>'
      ],
      function (expect, subject, query) {
        if (expect.flags.no || expect.flags.not) {
          return expect(subject.querySelectorAll(query), 'to satisfy', []);
        }

        expect.subjectOutput = function (output) { return expect.inspect(subject, Infinity, output); };

        return expect(subject.querySelectorAll(query), 'not to satisfy', []);
      }
    );

    expect.exportAssertion(
      '<DOMDocument|DOMElement|DOMDocumentFragment> [not] to contain test id <string>',
      function (expect, subject, testId) {
        expect.errorMode = 'nested';

        var escapedTestId = JSON.stringify(testId);

        return expect(
          subject,
          '[not] to contain elements matching',
          ("[data-test-id=" + escapedTestId + "]")
        );
      }
    );

    expect.exportAssertion(
      '<DOMDocument|DOMElement|DOMDocumentFragment> [not] to match <string>',
      function (expect, subject, query) {
        expect.subjectOutput = function (output) { return expect.inspect(subject, Infinity, output); };

        return expect(matchesSelector(subject, query), '[not] to be true');
      }
    );

    expect.exportAssertion(
      '<DOMDocument|DOMElement|DOMDocumentFragment> [not] to have test id <string>',
      function (expect, subject, testId) {
        expect.errorMode = 'nested';
        expect.subjectOutput = function (output) { return expect.inspect(subject, Infinity, output); };

        var escapedTestId = JSON.stringify(testId);

        return expect(
          subject,
          '[not] to match',
          ("[data-test-id=" + escapedTestId + "]")
        );
      }
    );

    expect.exportAssertion(
      '<string> [when] parsed as (html|HTML) [fragment] <assertion?>',
      function (expect, subject) {
        expect.errorMode = 'nested';
        return expect.shift(parseHtml(subject, expect.flags.fragment));
      }
    );

    expect.exportAssertion(
      '<string> [when] parsed as (xml|XML) <assertion?>',
      function (expect, subject) {
        expect.errorMode = 'nested';
        return expect.shift(parseXml(subject));
      }
    );

    function memoize(fn) {
      var map = new Map();
      return function (node) {
        var spec = map.get(node);
        if (typeof spec === 'undefined') {
          spec = fn(node);
          map.set(node, spec);
        }
        return spec;
      };
    }

    function scoreElementAgainstSpec(
      element,
      spec,
      memoizedConvertDOMNodeToSatisfySpec
    ) {
      var isTextSimilar = function (value, valueSpec) {
        var actual = (value || '').trim().toLowerCase();
        if (typeof valueSpec === 'string') {
          if (actual === valueSpec.trim().toLowerCase()) {
            return true;
          }
        } else if (valueSpec instanceof RegExp) {
          if (valueSpec.test(actual)) {
            return true;
          }
        } else if (typeof valueSpec === 'function') {
          return true;
        }

        return false;
      };

      var isHtml = isInsideHtmlDocument(element);

      var score = 0;

      var nodeName = isHtml
        ? element.nodeName.toLowerCase()
        : element.nodeName;

      if (isTextSimilar(nodeName, spec.name)) {
        score++;
      }

      if (isTextSimilar(element.textContent, spec.textContent)) {
        score++;
      }

      if (typeof element.hasAttribute === 'function') {
        var attributes = spec.attributes || {};
        var className = attributes.class;
        var style = attributes.style;

        if (className && element.hasAttribute('class')) {
          if (typeof className === 'string') {
            var expectedClasses = getClassNamesFromAttributeValue(className);
            var actualClasses = getClassNamesFromAttributeValue(
              element.getAttribute('class')
            );

            expectedClasses.forEach(function (expectedClass) {
              if (actualClasses.indexOf(expectedClass) !== -1) {
                score++;
              }
            });
          } else if (isTextSimilar(element.getAttribute('class'), className)) {
            score++;
          }
        }

        if (style && element.hasAttribute('style')) {
          var expectedStyles =
            typeof style === 'string' ? styleStringToObject(style) : style;
          var actualStyles = styleStringToObject(
            element.getAttribute('style')
          );

          Object.keys(expectedStyles).forEach(function (styleName) {
            var expectedStyle = expectedStyles[styleName];
            var actualStyle = actualStyles[styleName];

            if (actualStyle) {
              score++;
            }

            if (isTextSimilar(actualStyle, expectedStyle)) {
              score++;
            }
          });
        }

        var specialAttributes = ['style', 'class'];
        var ids = ['id', 'data-test-id', 'data-testid'];

        Object.keys(attributes).forEach(function (attributeName) {
          if (specialAttributes.indexOf(attributeName) !== -1) {
            return; // skip
          }

          if (element.hasAttribute(attributeName)) {
            if (typeof attributes[attributeName] === 'boolean') {
              score++;
            }

            if (
              element.getAttribute(attributeName) === attributes[attributeName]
            ) {
              score += ids.indexOf(attributeName) === -1 ? 1 : 100;
            }
          } else if (typeof attributes[attributeName] === 'undefined') {
            score++;
          }
        });
      }

      var expectedChildren = spec.children || [];

      expectedChildren.forEach(function (childSpec, i) {
        var child = element.childNodes[i];
        var childType = expect.findTypeOf(child);
        if (expect.findTypeOf(childSpec).is('DOMNode')) {
          childSpec = memoizedConvertDOMNodeToSatisfySpec(childSpec);
        }

        if (!child) {
          return;
        }

        if (typeof childSpec.nodeType === 'number') {
          if (child.nodeType === childSpec.nodeType) {
            if (childType.is('DOMElement')) {
              // Element
              score += scoreElementAgainstSpec(
                element.childNodes[i],
                memoizedConvertDOMNodeToSatisfySpec(childSpec),
                memoizedConvertDOMNodeToSatisfySpec
              );
            }

            score++;
          } else if (expect.findTypeOf(childSpec).is('DOMIgnoreComment')) {
            score++;
          }
        } else if (
          childType.is('DOMElement') &&
          typeof childSpec === 'object'
        ) {
          score += scoreElementAgainstSpec(
            element.childNodes[i],
            childSpec,
            memoizedConvertDOMNodeToSatisfySpec
          );
        } else if (
          childType.is('DOMTextNode') &&
          isTextSimilar(child.nodeValue, childSpec)
        ) {
          score++;
        }
      });

      return score;
    }

    function findMatchesWithGoodScore(
      data,
      spec,
      memoizedConvertDOMNodeToSatisfySpec
    ) {
      memoizedConvertDOMNodeToSatisfySpec =
        memoizedConvertDOMNodeToSatisfySpec ||
        memoize(convertDOMNodeToSatisfySpec);
      var elements =
        typeof data.length === 'number'
          ? Array.prototype.slice.call(data)
          : [data];

      var result = [];
      var bestScore = 0;

      elements.forEach(function (element) {
        var score = scoreElementAgainstSpec(
          element,
          spec,
          memoizedConvertDOMNodeToSatisfySpec
        );
        bestScore = Math.max(score, bestScore);

        if (score > 0 && score >= bestScore) {
          result.push({ score: score, element: element });
        }

        for (var i = 0; i < element.childNodes.length; i += 1) {
          var child = element.childNodes[i];
          if (child.nodeType === 1) {
            result.push.apply(
              result, findMatchesWithGoodScore(
                child,
                spec,
                memoizedConvertDOMNodeToSatisfySpec
              )
            );
          }
        }
      });

      result.sort(function (a, b) { return b.score - a.score; });

      if (result.length > 0) {
        var bestScore$1 = result[0].score;

        return result.filter(function (ref) {
          var score = ref.score;

          return score === bestScore$1;
        });
      }

      return result;
    }

    expect.exportAssertion(
      '<DOMDocument|DOMElement|DOMDocumentFragment|DOMNodeList> [not] to contain <DOMElement|object|string>',
      function (expect, subject, value) {
        var nodes = subject.childNodes || makeAttachedDOMNodeList(subject);
        var isHtml = isInsideHtmlDocument(
          subject.childNodes ? subject : nodes
        );
        var valueType = expect.findTypeOf(value);
        var spec = value;

        if (valueType.is('expect.it')) {
          throw new Error(
            'Unsupported value for "to contain" assertion: expect.it'
          );
        } else if (valueType.is('DOMElement')) {
          spec = convertDOMNodeToSatisfySpec(value, isHtml);
        } else if (valueType.is('string')) {
          var documentFragment = isHtml
            ? parseHtml(value, true)
            : parseXml(value);

          if (documentFragment.childNodes.length !== 1) {
            throw new Error(
              'HTMLElement to contain string: Only a single node is supported'
            );
          }

          spec = convertDOMNodeToSatisfySpec(
            documentFragment.childNodes[0],
            isHtml
          );

          if (typeof spec === 'string') {
            throw new Error(
              'HTMLElement to contain string: please provide a HTML structure as a string'
            );
          }

          expect.argsOutput = function (output) { return output.appendInspected(documentFragment.childNodes[0]); };

          ensureSupportedSpecOptions(spec);
        }

        var scoredElements = findMatchesWithGoodScore(nodes, spec);

        if (expect.flags.not) {
          if (scoredElements.length > 0) {
            return expect.withError(
              function () { return expect(
                  scoredElements.map(function (ref) {
                    var element = ref.element;

                    return element;
                  }),
                  'not to have an item satisfying',
                  spec
                ); },
              function () {
                var bestMatch = scoredElements[0].element;

                expect.subjectOutput = function (output) { return expect.inspect(subject, Infinity, output); };

                expect.fail({
                  diff: function (output, diff, inspect, equal) {
                    return output
                      .error('Found:')
                      .nl(2)
                      .appendInspected(bestMatch);
                  }
                });
              }
            );
          }
        } else {
          if (scoredElements.length === 0) {
            expect.subjectOutput = function (output) { return expect.inspect(subject, Infinity, output); };
            expect.fail();
          }

          return expect.withError(
            function () { return expect(
                scoredElements.map(function (ref) {
                  var element = ref.element;

                  return element;
                }),
                'to have an item satisfying',
                spec
              ); },
            function () {
              var bestMatch = scoredElements[0].element;

              return expect(bestMatch, 'to satisfy', spec);
            }
          );
        }
      }
    );
  }
};
